@@ -37,8 +37,8 @@ TextLine::TextLine(int attribute, int context)
 }
 
 TextLine::~TextLine() {
-  delete text;
-  delete attribs;
+  delete[] text;
+  delete[] attribs;
 }
 
 void TextLine::move(int pos, int n) {
@@ -162,9 +162,9 @@ void TextLine::unWrap(TextLine *nextLine, int pos) {
   nextLine->del(0,pos);
 }
 
-void TextLine::removeSpaces() {
+void TextLine::removeTrailingWhitespace() {
 
-  while (len > 0 && text[len - 1] == ' ') len--;
+  while (len > 0 && text[len - 1] <= 32) len--;
 }
 
 int TextLine::firstChar() {
@@ -185,6 +185,27 @@ int TextLine::lastChar() {
   return (z > 0) ? z : -1;
 }
 
+int TextLine::indentTabs() {
+  int z;
+
+  z = 0;
+  while (z < len && (unsigned char) text[z] == '\t')
+    z++;
+  return z;
+}
+
+int TextLine::indentSpaces() {
+  int y, z;
+
+  z = indentTabs();
+  y = 0;
+  while (z < len && (unsigned char) text[z] <= 32) {
+    z++; // skip ws after tabs
+    y++;
+  }
+  return y;
+}
+
 char TextLine::getChar( int pos ) const {
 
   if (pos >= 0 && pos < len )
@@ -239,7 +260,6 @@ const char* TextLine::getText() {
   return text;
 }
 
-
 void TextLine::select(bool sel, int start, int end) {
   int z;
 
@@ -360,15 +380,15 @@ if (!newtext || !newattribs) {
 }
     memcpy(newtext,text,len);
     memcpy(newattribs,attribs,len);
-    delete text;
-    delete attribs;
+    delete[] text;
+    delete[] attribs;
     text = newtext;
     attribs = newattribs;
   }
 }
 
 
-Attribute::Attribute() : font(), fm(font) {
+Attribute::Attribute() : font(), printFont(), fm(font), printFM(printFont) {
 }
 
 void Attribute::setFont(const QFont &f) {
@@ -376,6 +396,11 @@ void Attribute::setFont(const QFont &f) {
   fm = QFontMetrics(f);
 }
 
+void Attribute::setPrintFont(const QFont &f) {
+  printFont = f;
+  printFM = QFontMetrics(f);
+}
+
 
 KWAction::KWAction(Action a, PointStruc &aCursor)
   : action(a), cursor(aCursor), len(0), text(0L), textLen(0) {
@@ -422,11 +447,17 @@ KWriteDoc::KWriteDoc(HlManager *hlManager, const char *path)
   : QObject(0L)
   ,hlManager(hlManager)
   ,highlight(0L)
+  ,commentAttribute(-1)
   ,kWriteConfig(0L)
   ,tabChars(8)
   ,tabWidth(0)
+  ,printTabWidth(0)
   ,fontHeight(0)
+  ,printFontHeight(0)
   ,fontAscent(0)
+  ,printFontAscent(0)
+  ,indentLength(2)
+  ,bwPrinting(true)
   ,newDocGeometry(false)
   ,longestLine(0L)
   ,maxLength(0)
@@ -563,6 +594,7 @@ void KWriteDoc::readConfig(KConfig *config) {
 
   setTabWidth(config->readNumEntry("TabWidth",8));
   setIndentLength(config->readNumEntry("IndentLength",2));
+  setBWPrinting(config->readBoolEntry("BlackAndWhitePrinting",true));
   setUndoSteps(config->readNumEntry("UndoSteps",5000));
   for (z = 0; z < 5; z++) {
     sprintf(s,"Color%d",z);
@@ -576,6 +608,7 @@ void KWriteDoc::writeConfig(KConfig *config) {
 
   config->writeEntry("TabWidth",tabChars);
   config->writeEntry("IndentLength",indentLength);
+  config->writeEntry("BlackAndWhitePrinting",bwPrinting);
   config->writeEntry("UndoSteps",undoSteps);
   for (z = 0; z < 5; z++) {
     sprintf(s,"Color%d",z);
@@ -1058,11 +1091,91 @@ void KWriteDoc::insertChar(KWriteView *view, VConfig &c, char ch) {
   recordEnd(view,c);
 }
 
+// Here's the indentation behaviour so that everybody knows
+// what the code is supposed to do.
+
+// "enter indents" -- special behaviour of newline
+// 1. generally, the new line is filled with same whitespace as
+//    the first non-empty line above it
+// 2. if the current cursor is enclosed by the brace pair,
+//    the closing brace gets delegated to a new line of its own
+//    TODO: check for "normal text" attribute, do not indent otherwise
+// 3. if either the last character, or a character preceding the cursor
+//    is an opening brace, the new line is indented a level deeper
+//    TODO: check for "normal text" attribute, do not indent otherwise
+// 4. TODO: if a parenthesis with "normal text" attribute is opened,
+//    and not closed, newline will indent to one cell to the right
+//    of the opened parenthesis -- say continuation of if (a || <newline> ...
+// (1) is enabled by cfAutoIndent
+// (2,3) are enabled by cfIndentBraces
+// (4) is enabled by cfIndentParentheses
+
+// "tab indents" -- special behaviour of tab & shift-tab
+// 0. if cursor is:
+//    - past end of text in a non-empty line,
+//    - in a (possibly zero-length) whitespace before // or /*, but not in the
+//      whitespace at beginning of the line (so that we can normally indent
+//      multiline comments),
+//    insert regular tab and don't indent (and the algorithm is finished)
+// 1. treat the first non-empty line above the current one as baseline
+// 2. if the current line has closing brace "}" as its first character,
+//    then decrease the baseline indentation by one level
+// (tab)
+// 3. if the current line has less indentation than the baseline,
+//    replicate the baseline indentation here
+// 4. if the current line has same or more indentation than the
+//    baseline, add one indentation level
+// (3,4) are enabled by cfTabIndent
+// (shift-tab)
+// 3. if the current line has more indentation than the baseline,
+//    replicate the baseline indentation here
+// 4. if the current line hase same or less indentation than the
+//    baseline, remove one indentation level
+// (3,4) are enabled by cfTabIndent
+
+// "backspace (un)indents" -- special behaviour of backspace before text
+// 1. set indentation to that of the first non-empty line above
+//    it that has less indentation than the current line
+// (1) is enabled by cfBackspaceIndent
+
+// NB. 2.1's behaviour was (somewhat) working due to lucky circumstances.
+// I dream that this implementation works because it was meant to ;-)
+
+// -- The (hopefully not) mess in newLine() and tab() below is
+//    courtesy of Kuba Ober (kuba@mareimbrium.org), yell at him
+//    if you have flames / ideas / bugreports.
+
 #ifdef DEBUG
 #include <iostream>
 #endif
 
+int KWriteDoc::seekIndentRef(QList<TextLine> & contents, int & tabs, int & spaces) {
+  int cells = 0;
+  TextLine * textLine = contents.current();
+  while (textLine != 0) {
+    tabs = textLine->indentTabs();
+    spaces = textLine->indentSpaces();
+    cells = tabs * tabChars + spaces;
+    if (cells > 0 || (cells == 0 && textLine->firstChar() == 0)) // we've got our position so we're done
+      break;
+    textLine = contents.prev();
+  };
+  return cells;
+}
+
+int KWriteDoc::seekIndentRef(QList<TextLine> & contents) {
+  int tabs;
+  int spaces;
+  return seekIndentRef(contents, tabs, spaces);
+}
+
 void KWriteDoc::newLine(KWriteView *view, VConfig &c) {
+  // NB: silent assumption is that what indentSpaces() counts
+  // indeed consists of all spaces. If somebody types tab, tab,
+  // space, tab, text, and tabs are *not* indents that would be
+  // replaced with spaces, then this will be borken as it used to
+  // be in 2.1-release, but such an antisocial behaviour should be
+  // damned anyway ;-)) Kuba
 
   recordStart(c.cursor);
 
@@ -1072,122 +1185,271 @@ void KWriteDoc::newLine(KWriteView *view, VConfig &c) {
     c.cursor.y++;
     c.cursor.x = 0;
   } else {
-    // indent the new line
+    // insert new line and indent it
 
-    // find the column to indent to
     TextLine* textLine = contents.at(c.cursor.y); // check the current line
     // NB: accessing the list like that is expensive
 
-    // compute where to indent to
-    int pos = textLine->firstChar(); // first non-ws char
+    // check if the user wants to insert a line before this one
     bool prevLine = false;
-    if (pos > c.cursor.x) {  // what the user wants is to
-      c.cursor.x = pos;     // insert a line before this one!
+    if (textLine->firstChar() > c.cursor.x) {
+      c.cursor.x = textLine->firstChar();
       prevLine = true;
     }
-    // seek a first char in previous lines
-    do {
-      pos = textLine->firstChar();
-      if (!prevLine && textLine->getChar(textLine->lastChar())=='{')
-        // opening brace
-        pos += indentLength;
-      if (pos >= 0) // we've got our position
-        break;     // terminate loop
-      textLine = contents.prev();
-    } while (textLine);
 
-    // insert the newline and indent the cursor
+    // find out current indentation level --
+    // seek the first char in this first non-empty line, starting at
+    // current line
+    int tabs = 0; // how many tabs were used in indentation found
+    int spaces = 0; // how many spaces were used in indentation found
+    int cells = 0; // how many character cells were used (ie. accounting tab width)
+
+    cells = seekIndentRef(contents, tabs, spaces);
+
+    // check if we want to newline after an opening brace,
+    // and if there's a closing brace following the opening one
+    // (i.e. if auto brackets were used)
+    bool openingBrace = false;
+    bool closingBrace = false;
+
+    if (!prevLine && (c.flags & cfIndentBraces)) {
+      int lastChar = textLine->lastChar();
+      openingBrace = (c.cursor.x > lastChar && textLine->getChar(lastChar)=='{') ||
+                     (c.cursor.x > 0 && textLine->getChar(c.cursor.x - 1)=='{');
+      closingBrace = openingBrace &&
+                     (c.cursor.x > 0) &&
+                     (textLine->getChar(c.cursor.x)=='}');
+    }
+
+    // seek to the line with opening parenthesis if
+    // there's a closing parenthesis in the current line,
+    // like this:
+    // <spc><spc>if (a ||              <-- get indentation from here!
+    // <spc><spc><spc><spc><spc><spc>b) {
+    // TODO: we assume that the closing parenthesis is just one
+    // character away, this may not be true -- right now I'm a coward ;-) - Kuba
+
+    if (c.cursor.x > 2 && textLine->getChar(c.cursor.x - 3) == ')') {
+       BracketMark bm;
+       PointStruc csr(c.cursor.x - 2, c.cursor.y);
+       #ifdef DEBUG
+       std::cerr << "line " << c.cursor.y << ", will indent using line with matching ()" << std::endl;
+       #endif
+       newBracketMark(csr, bm);
+       // if matching bracket is found, get the indentation
+       // depth from that line
+       if (bm.eXPos >= 0) {
+         #ifdef DEBUG
+         std::cerr << "matched bracket at line: " << bm.cursor.y << std::endl;
+         #endif
+         TextLine* textLine = contents.at(bm.cursor.y);
+         spaces = textLine->indentSpaces();
+         tabs = textLine->indentTabs();
+         cells = tabs * tabChars + spaces;
+       }
+    }
+
+    // insert the newline
     recordAction(KWAction::newLine,c.cursor); // insert and record newline
     c.cursor.y++;
     c.cursor.x = 0;
-    if (pos > 0) {
-      char* buf = new char[pos];
-      memset(buf, ' ', pos);
+
+    // if there is a closing brace, insert another newline and indent
+    if (closingBrace) {
+      if (c.flags & cfReplaceTabs) {
+        tabs = 0;
+        spaces = cells;
+      }
+
+      recordAction(KWAction::newLine,c.cursor);
+      c.cursor.y++;
+
+      if (cells > 0) {
+        char* buf = new char[tabs + spaces];
+        if (tabs > 0) memset(buf, '\t', tabs);
+        if (spaces > 0) memset(buf + tabs, ' ', spaces);
+        PointStruc linebeg(0, c.cursor.y);
+        recordInsert(linebeg, &buf[0], tabs + spaces);
+      }
+
+      c.cursor.y--;
+    }
+
+    // increase indentation if we're after a brace
+    if (openingBrace) {
+      cells += indentLength;
+      tabs = cells / tabChars;
+      spaces = cells % tabChars;
+    }
+
+    // indent the line where the cursor ends up
+    if (cells > 0) {
+      if (c.flags & cfReplaceTabs) {
+        tabs = 0;
+        spaces = cells;
+      }
+
+      char* buf = new char[tabs + spaces];
+      if (tabs > 0) memset(buf, '\t', tabs);
+      if (spaces > 0) memset(buf + tabs, ' ', spaces);
       PointStruc linebeg(0, c.cursor.y);
-      recordInsert(linebeg, &buf[0], pos);
-      c.cursor.x = pos;
+      recordInsert(linebeg, &buf[0], tabs + spaces);
+      delete[] buf;
+      c.cursor.x = tabs + spaces;
     }
   }
 
   recordEnd(view,c);
 }
 
-void KWriteDoc::tab(KWriteView *view, VConfig &c) {
+// returns the smallest non-zero value of the pair (a, b), or zero if both are zero
+template <typename T>
+static inline T minnz(T a, T b) { return (((a != 0) && (b != 0) && (a < b)) || (b == 0)) ? a : b; }
+// returns the largest of the pair (a, b), or a if they are same
+template <typename T>
+static inline T max(T a, T b) { return (a >= b) ? a : b; }
+
+void KWriteDoc::commonTab(KWriteView * view, VConfig & c, bool add) {
+
+  // TODO: it would be cleaner to put it where it belongs, that is at all points
+  // where highlight items can get changed
+  if (commentAttribute == -1) {
+    ItemDataList idl;
+    ItemDataList & list = idl;
+    highlight->getItemDataList(list);
+    #ifdef DEBUG
+    std::cerr << "highlight->getItemDataList(list); list.count() = " << list.count() << std::endl;
+    #endif
+    for (ItemData * item = list.first(); item != 0; item = list.next()) {
+      if (item->name == "Comment") commentAttribute = list.at();
+    }
+    #ifdef DEBUG
+    std::cerr << "commentAttribute = " << commentAttribute << std::endl;
+    #endif
+  }
 
   recordStart(c.cursor);
 
-  if (!(c.flags & cfTabIndent)) {
-    // auto indentation hasn't been chosen
+  TextLine* textLine = contents.at(c.cursor.y);
+  // NB: accessing the list like that is expensive
+
+  const char * curline = textLine->getString();
+  const char * curtext;
+  const char * comment;
+  const char * dummy = "abc"; // dummy text to avoid excess test-for-NULL logic
+  curtext = (textLine->length() > c.cursor.x) ? curline + c.cursor.x : dummy + 1;
+  comment = minnz(strstr(curtext, "//"), strstr(curtext, "/*"));
+  comment = (comment != 0) ? comment : dummy;
+
+  if ((!(c.flags & cfTabIndent)) || // tab doesn't indent
+      ((textLine->lastChar() > 0) && (c.cursor.x > textLine->lastChar())) || // past end of line
+      ((strspn(curtext, " \t") == (size_t)(comment - curtext)) && // in the whitespace before comment
+       (strspn(curline, " \t") < (size_t)(comment - curline))) || // but not in leading whitespace
+      (textLine->getAttr(max(c.cursor.x - 1, 0)) == commentAttribute)) { // inside of comment
     // insert regular tab like we know it
-    insertChar(view, c, '\t');
+    if (add) insertChar(view, c, '\t');
   } else {
     // indent the new line
 
-    // find the column to indent to
-    TextLine* textLine = contents.at(c.cursor.y); // check the current line
-    // NB: accessing the list like that is expensive
+    // find out baseline indentation level
+    bool preserve = true; // shall we preserve original indentation makeup?
+    int tabs = 0; // how many tabs were used in indentation found
+    int spaces = 0; // how many spaces were used in indentation found
+    int cells = 0; // how many character cells were used (ie. accounting tab width)
+   
+    if (c.cursor.y > 0) {
+      textLine = contents.at(c.cursor.y - 1); // check the previous line
 
-    // compute where to indent to
-    bool setCursor = false;
-    int indentPos = 0;		// first non-ws char
-    int indent = 0;		// how many chars to indent
+      cells = seekIndentRef(contents, tabs, spaces);
+
+      textLine = contents.current();
+
+      // adjust indentation levels for opening brace in this line
+      if (textLine->getChar(textLine->lastChar())=='{') { // opening brace
+        cells += indentLength;
+        preserve = false; // preserving spaces only make sense to align (...) stuff --
+                          // opening brace surely means the parentheses were closed
+      }
+    }
+
+    textLine = contents.at(c.cursor.y);
+
+    // adjust baseline for closing brace in this line
     if (textLine->getChar(textLine->firstChar())=='}') {
-      // closing brace
-      indent -= indentLength;
-      setCursor = true;
-    }
-    // seek a first char in previous lines
-    textLine = contents.prev();
-    while (textLine) {
-      indentPos = textLine->firstChar();
-      if (textLine->getChar(textLine->lastChar())=='{') // opening brace
-        indent += indentLength;
+      cells -= indentLength;
+      if (cells < 0) cells = 0;
       #ifdef DEBUG
-      std::cerr << "indent pos=" << indentPos << std::endl;
+      std::cerr << "adjusted (}) to cells=" << cells << std::endl;
       #endif
-      if (indentPos >= 0) // we've got our position
-        break;     // terminate loop
-      textLine = contents.prev();
     }
-    indentPos += indent;
 
+    // get current indentation level
+    int curCells = textLine->indentTabs() * tabChars + textLine->indentSpaces();
+
+    // select desired indentation
+    int indentPos = 0;
+    if (add) {
+      if (curCells < cells)
+        indentPos = cells; // at least the baseline
+      else
+        indentPos = curCells + indentLength; // allow indenting past it
+    } else {
+      if (curCells > cells)
+        indentPos = cells; // at most the baseline
+      else {
+        indentPos = curCells - indentLength; // allow un-indenting past it
+        if (indentPos < 0) indentPos = 0;
+      }
+    }
+
+    #ifdef DEBUG
+    std::cerr << "curCells " << curCells << " cells " << cells;
+    std::cerr << " indentPos " << indentPos << std::endl;
+    #endif
+
+    // alter reference line indentation makeup only if changing indentation depth
+    preserve &= cells == indentPos;
+    if (! preserve) {
+      tabs = indentPos / tabChars;
+      spaces = indentPos % tabChars;
+    }
+
+    if (c.flags & cfReplaceTabs) {
+      tabs = 0;
+      spaces = indentPos;
+    }
+
+    // reindent
+    int remove = (textLine->firstChar() >= 0) ? textLine->firstChar() : textLine->length();
+
+    char * buf = new char[tabs + spaces];
     if (indentPos > 0) {
-
-      // if the line does not start on indent pos align it there
-      textLine = contents.at(c.cursor.y);
-      int curPos = textLine->firstChar();
-      if (curPos < indentPos) {
-        // indent
-        int len = indentPos-curPos;
-        char* buf = new char[len];
-        memset(buf, ' ', len);
-        PointStruc linebeg(0, c.cursor.y);
-        recordInsert(linebeg, &buf[0], len);
-        c.cursor.x += len;
-        delete[] buf;
-      }
-      else if (curPos > indentPos) {
-        // unindent
-        PointStruc beg(indentPos, c.cursor.y);
-        recordDelete(beg, curPos-indentPos);
-        c.cursor.x -= curPos-indentPos;
-      }
-
-      if (c.cursor.x < curPos)
-      // if cursor before the first char then align cursor
-      // with the start of indentation
-        setCursor = true;
+      if (tabs > 0) memset(buf, '\t', tabs);
+      if (spaces > 0) memset(buf + tabs, ' ', spaces);
     }
 
-    if (setCursor) {
-      c.cursor.x = indentPos;
-    }
+    PointStruc beg(0, c.cursor.y);
+    recordReplace(beg, remove, &buf[0], tabs + spaces);
 
+    delete[] buf;
+
+    // restore cursor position
+    c.cursor.x += tabs + spaces;
+    if (remove > 0)
+      c.cursor.x -= remove;
   }
 
   recordEnd(view,c);
 }
 
+void KWriteDoc::tab(KWriteView *view, VConfig &c) {
+  commonTab(view, c, true);
+}
+
+void KWriteDoc::shiftTab(KWriteView *view, VConfig &c) {
+  commonTab(view, c, false);
+}
+
 void KWriteDoc::killLine(KWriteView *view, VConfig &c) {
 
   recordStart(c.cursor);
@@ -1321,6 +1583,28 @@ void KWriteDoc::updateFontData() {
   int i, z;
   KWriteView *view;
 
+  // do calculations for printer
+  maxAscent = 0;
+  maxDescent = 0;
+  minTabWidth = 0xffffff;
+  maxTabWidth = 0;
+
+  for (z = 0; z < nAttribs; z++) {
+    i = attribs[z].printFM.ascent();
+    if (i > maxAscent) maxAscent = i;
+    i = attribs[z].printFM.descent();
+    if (i > maxDescent) maxDescent = i;
+    i = attribs[z].printFM.width('x');
+    if (i < minTabWidth) minTabWidth = i;
+    if (i > maxTabWidth) maxTabWidth = i;
+  }
+
+  printFontHeight = maxAscent + maxDescent + 1;
+  printFontAscent = maxAscent;
+  printTabWidth = tabChars*(maxTabWidth + minTabWidth)/2;
+
+  
+  // do calculations for screen
   maxAscent = 0;
   maxDescent = 0;
   minTabWidth = 0xffffff;
@@ -1375,7 +1659,13 @@ void KWriteDoc::setIndentLength(int length){
   indentLength = length;
 }
 
-void KWriteDoc::updateLines(int startLine, int endLine, int flags)
+/** Set black and white printing width */
+void KWriteDoc::setBWPrinting(bool bw)
+{
+  bwPrinting=bw;
+}
+
+void KWriteDoc::updateLines(int startLine, int curLine, int endLine, int flags)
 {
   TextLine *textLine;
   int line, lastLine;
@@ -1395,8 +1685,8 @@ void KWriteDoc::updateLines(int startLine, int endLine, int flags)
     textLine = contents.at(line);
     if (line <= endLine)
     {
-      if (flags & cfRemoveSpaces)
-        textLine->removeSpaces();
+      if ((flags & cfRemoveSpaces) && (line != curLine))
+        textLine->removeTrailingWhitespace();
       updateMaxLength(textLine);
     }
 
@@ -2141,172 +2431,18 @@ QColor &KWriteDoc::cursorCol(int x, int y) {
     return a->col;
 }
 
-/*
-void KWriteDoc::paintTextLine(QPainter &paint, int line,
-                              int xPos, int xStart, int xEnd, int yPos) {
-  int y;
-  TextLine *textLine;
-  int z, x;
-  char ch;
-  Attribute *a = 0;
-  int attr, nextAttr;
-  int xs;
-  int xc, zc;
-
-  y = line*fontHeight - yPos;
-  if (line >= (int) contents.count()) {
-    paint.fillRect(xStart - xPos,y,xEnd - xStart,fontHeight,colors[4]);
-    return;
-  }
-//printf("xStart = %d, xEnd = %d, line = %d\n",xStart,xEnd,line);
-//printf("text = ");
-  textLine = contents.at(line);
-
-  z = 0;
-  x = 0;
-  do {
-    xc = x;
-    ch = textLine->getChar(z);
-    if (ch == '\t') {
-      x += tabWidth - (x % tabWidth);
-    } else {
-      a = attribs[textLine->getAttr(z)];
-      x += a->fm.width(&ch,1);
-    }
-    z++;
-  } while (x <= xStart);
-  zc = z - 1;
-
-  xs = xStart;
-  attr = textLine->getRawAttr(zc);
-  while (x < xEnd) {
-    nextAttr = textLine->getRawAttr(z);
-    if ((nextAttr ^ attr) & taSelectMask) {
-      paint.fillRect(xs - xPos,y,x - xs,fontHeight,colors[attr >> taShift]);
-      xs = x;
-      attr = nextAttr;
-    }
-    ch = textLine->getChar(z);
-    if (ch == '\t') {
-      x += tabWidth - (x % tabWidth);
-    } else {
-      a = attribs[attr & taAttrMask];
-      x += a->fm.width(&ch,1);
-    }
-    z++;
-  }
-  paint.fillRect(xs - xPos,y,xEnd - xs,fontHeight,colors[attr >> taShift]);
-//int len = textLine->length();
-  y += fontAscent -1;
-  attr = -1;
-  while (xc < xEnd) {
-    ch = textLine->getChar(zc);
-    if (ch == '\t') {
-      xc += tabWidth - (xc % tabWidth);
-    } else {
-      nextAttr = textLine->getRawAttr(zc);
-      if (nextAttr != attr) {
-        attr = nextAttr;
-        a = attribs[attr & taAttrMask];
-        if (attr & taSelectMask) paint.setPen(a->selCol); else paint.setPen(a->col);
-        paint.setFont(a->font);
-      }
-      paint.drawText(xc - xPos,y,&ch,1);
-      xc += a->fm.width(&ch,1);
-//if (zc < len) printf("%c",ch);
-    }
-    zc++;
-  }
-
-//printf("\n");
+void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, bool showTabs, bool printing)
+{
+    paintTextLine(paint, line, 0, xStart, xEnd, showTabs, printing);
 }
-*/
-
-// void KWriteDoc::paintTextLine(QPainter &paint, int line,
-//                               int xStart, int xEnd) {
-//   int y;
-//   TextLine *textLine;
-//   int z, x;
-//   char ch;
-//   Attribute *a = 0L;
-//   int attr, nextAttr;
-//   int xs;
-//   int xc, zc;
-
-//   y = 0;//line*fontHeight - yPos;
-//   if (line >= (int) contents.count()) {
-//     paint.fillRect(0,y,xEnd - xStart,fontHeight,colors[4]);
-//     return;
-//   }
-// //printf("xStart = %d, xEnd = %d, line = %d\n",xStart,xEnd,line);
-// //printf("text = ");
-//   textLine = contents.at(line);
-
-//   z = 0;
-//   x = 0;
-//   do {
-//     xc = x;
-//     ch = textLine->getChar(z);
-//     if (ch == '\t') {
-//       x += tabWidth - (x % tabWidth);
-//     } else {
-//       a = &attribs[textLine->getAttr(z)];
-//       x += a->fm.width(&ch,1);
-//     }
-//     z++;
-//   } while (x <= xStart);
-//   zc = z - 1;
-
-//   xs = xStart;
-//   attr = textLine->getRawAttr(zc);
-//   while (x < xEnd) {
-//     nextAttr = textLine->getRawAttr(z);
-//     if ((nextAttr ^ attr) & (taSelectMask | 256)) {
-//       paint.fillRect(xs - xStart,y,x - xs,fontHeight,colors[attr >> taShift]);
-//       xs = x;
-//       attr = nextAttr;
-//     }
-//     ch = textLine->getChar(z);
-//     if (ch == '\t') {
-//       x += tabWidth - (x % tabWidth);
-//     } else {
-//       a = &attribs[attr & taAttrMask];
-//       x += a->fm.width(&ch,1);
-//     }
-//     z++;
-//   }
-//   paint.fillRect(xs - xStart,y,xEnd - xs,fontHeight,colors[attr >> taShift]);
-// //int len = textLine->length();
-//   y += fontAscent -1;
-//   attr = -1;
-//   while (xc < xEnd) {
-//     ch = textLine->getChar(zc);
-//     if (ch == '\t') {
-//       xc += tabWidth - (xc % tabWidth);
-//     } else {
-//       nextAttr = textLine->getRawAttr(zc);
-//       if (nextAttr != attr) {
-//         attr = nextAttr;
-//         a = &attribs[attr & taAttrMask];
-//         if (attr & taSelectMask) paint.setPen(a->selCol); else paint.setPen(a->col);
-//         paint.setFont(a->font);
-//       }
-//       paint.drawText(xc - xStart,y,&ch,1);
-//       xc += a->fm.width(&ch,1);
-// //if (zc < len) printf("%c",ch);
-//     }
-//     zc++;
-//   }
 
-// //printf("\n");
-// }
 ////////////////////////////////////////////////////////////////////////
-void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, bool showTabs)
+void KWriteDoc::paintTextLine(QPainter &paint, int line, int y, int xStart, int xEnd, bool showTabs, bool printing)
 {
-  int y;
   TextLine* textLine;
   int len;
   const char *s;
+  QBrush whiteBrush(white);
   int z, x;
   char ch;
   Attribute *a = 0L;
@@ -2314,15 +2450,14 @@ void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, b
   int xs;
   int xc, zc;
 
-//  struct timeval tv1, tv2, tv3; //for profiling
-//  struct timezone tz;
-
-  y = 0;//line*fontHeight - yPos;
   if (line > lastLine()) {
-    paint.fillRect(0, y, xEnd - xStart,fontHeight, colors[4]);
+    if (printing)
+      paint.fillRect(0, y, xEnd - xStart, printFontHeight, whiteBrush);
+    else
+      paint.fillRect(0, y, xEnd - xStart, fontHeight, colors[4]);
     return;
   }
-
+  
   textLine = contents.at(line);
   if (!textLine) return;
   len = textLine->length();
@@ -2336,16 +2471,28 @@ void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, b
     zc = z;
     if (z == len) break;
     ch = s[z];//textLine->getChar(z);
-    if (ch == '\t') {
-      x += tabWidth - (x % tabWidth);
+    if (ch == '\t') 
+    {
+      if (printing)
+        x += printTabWidth - (x % printTabWidth);
+      else
+        x += tabWidth - (x % tabWidth);
+//      x += tabWidth - (x % tabWidth);
     } else {
       a = &attribs[textLine->getAttr(z)];
+    
+    if (printing)
+    {
+      a->printFM=paint.fontMetrics();
+      x += a->printFM.width(ch);//a->width(ch);
+    }
+    else  
       x += a->fm.width(ch);//a->width(ch);
+//      x += a->fm.width(ch);//a->width(ch);
     }
     z++;
   } while (x <= xStart);
 
-//gettimeofday(&tv1, &tz);
 
   // draw background
   xs = xStart;
@@ -2353,29 +2500,51 @@ void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, b
   while (x < xEnd) {
     nextAttr = textLine->getRawAttr(z);
     if ((nextAttr ^ attr) & (taSelectMask | 256)) {
-      paint.fillRect(xs - xStart, y, x - xs, fontHeight, colors[attr >> taShift]);
+      if (printing)
+        paint.fillRect(xs - xStart, y, x - xs, printFontHeight, whiteBrush);
+      else
+        paint.fillRect(xs - xStart, y, x - xs, fontHeight, colors[attr >> taShift]);
+//      paint.fillRect(xs - xStart, y, x - xs, fontHeight, colors[attr >> taShift]);
       xs = x;
       attr = nextAttr;
     }
     if (z == len) break;
     ch = s[z];//textLine->getChar(z);
     if (ch == '\t') {
-      x += tabWidth - (x % tabWidth);
-    } else {
+      if (printing)
+        x += printTabWidth - (x % printTabWidth);
+      else
+        x += tabWidth - (x % tabWidth);
+//      x += tabWidth - (x % tabWidth);
+    } else 
+    {
       a = &attribs[attr & taAttrMask];
+    if (printing)
+    {
+      a->printFM=paint.fontMetrics();
+      x += a->printFM.width(ch);//a->width(ch);
+    }
+    else  
       x += a->fm.width(ch);//a->width(ch);
+//      x += a->fm.width(ch);//a->width(ch);
     }
     z++;
   }
-  paint.fillRect(xs - xStart, y, xEnd - xs, fontHeight, colors[attr >> taShift]);
+  if (printing)
+    paint.fillRect(xs - xStart, y, x - xs, printFontHeight, whiteBrush);
+  else
+    paint.fillRect(xs - xStart, y, xEnd - xs, fontHeight, colors[attr >> taShift]);
+//  paint.fillRect(xs - xStart, y, xEnd - xs, fontHeight, colors[attr >> taShift]);
+  
   len = z; //reduce length to visible length
 
-//gettimeofday(&tv2, &tz);
-
-  // draw text
   x = xc;
   z = zc;
-  y += fontAscent;// -1;
+  if (printing)
+    y += printFontAscent;// -1;
+  else
+    y += fontAscent;// -1;
+// y += fontAscent;// -1;
   attr = -1;
   while (z < len) {
     ch = s[z];//textLine->getChar(z);
@@ -2383,7 +2552,14 @@ void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, b
       if (z > zc) {
         QString str=QString::fromLocal8Bit(&s[zc], z - zc /*+1*/);
         paint.drawText(x - xStart, y, str);
-        x += a->fm.width(str); //a->width(s);//&s[zc], z - zc);
+        if (printing)
+        {
+          a->printFM=paint.fontMetrics();
+          x += a->printFM.width(str);//a->width(ch);
+        }
+        else  
+          x  += a->fm.width(str);//a->width(ch);
+//       x += a->fm.width(str);//a->width(ch);
       }
       zc = z +1;
 
@@ -2393,33 +2569,61 @@ void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, b
           attr = nextAttr;
           a = &attribs[attr & taAttrMask];
 
-          if (attr & taSelectMask) paint.setPen(a->selCol);
-            else paint.setPen(a->col);
+          if (printing && bwPrinting)
+             paint.setPen(black);
+          else
+          {
+            if (attr & taSelectMask) paint.setPen(a->selCol);
+              else paint.setPen(a->col);
+          }
+        if (printing)
+          paint.setFont(a->printFont);
+        else
           paint.setFont(a->font);
+//          paint.setFont(a->font);
         }
 
-//        paint.drawLine(x - xStart, y -2, x - xStart, y);
-//        paint.drawLine(x - xStart, y, x - xStart + 2, y);
         paint.drawPoint(x - xStart, y);
         paint.drawPoint(x - xStart +1, y);
         paint.drawPoint(x - xStart, y -1);
       }
-      x += tabWidth - (x % tabWidth);
+      
+      if (printing)
+        x += printTabWidth - (x % printTabWidth);
+      else
+        x += tabWidth - (x % tabWidth);
+//      x += tabWidth - (x % tabWidth);
     } else {
       nextAttr = textLine->getRawAttr(z);
       if (nextAttr != attr) {
         if (z > zc) {
           QString str=QString::fromLocal8Bit(&s[zc], z - zc /*+1*/);
           paint.drawText(x - xStart, y, str);
-          x += a->fm.width(str);//a->width(s);//&s[zc], z - zc);
+          if (printing)
+          {
+             a->printFM=paint.fontMetrics();
+             x += a->printFM.width(str);//a->width(ch);
+          }
+          else  
+            x  += a->fm.width(str);//a->width(ch);
+//       x += a->fm.width(str);//a->width(ch);
           zc = z;
         }
         attr = nextAttr;
         a = &attribs[attr & taAttrMask];
 
-        if (attr & taSelectMask) paint.setPen(a->selCol);
-          else paint.setPen(a->col);
-        paint.setFont(a->font);
+        if (printing && bwPrinting)
+           paint.setPen(black);
+        else
+        {
+          if (attr & taSelectMask) paint.setPen(a->selCol);
+            else paint.setPen(a->col);
+        }
+        if (printing)
+          paint.setFont(a->printFont);
+        else
+          paint.setFont(a->font);
+//          paint.setFont(a->font);
       }
     }
     z++;
@@ -2428,210 +2632,12 @@ void KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd, b
     QString str=QString::fromLocal8Bit(&s[zc], z - zc /*+1*/);
     paint.drawText(x - xStart, y, str);
   }
-//gettimeofday(&tv3, &tz);
-//printf(" %d %d\n", tv2.tv_usec - tv1.tv_usec, tv3.tv_usec - tv2.tv_usec);
+
 }
 
 
-
-//void
-//KWriteDoc::paintTextLine(QPainter &paint, int line, int xStart, int xEnd)
-//{
-//  int y;
-//  TextLine *textLine;
-//  int z, x;
-//  char ch;
-//  Attribute *a = 0L;
-//  int attr, nextAttr;
-//  int xs;
-//  int xc, zc;
-//  static char aLB[256];
-//  int aCnt, xcp;
-//
-//  y = 0;//line*fontHeight - yPos;
-//  if (line >= (int) contents.count()) {
-//      paint.fillRect(0,y,xEnd - xStart,fontHeight,colors[4]);
-//      return;
-//  }
-////printf("xStart = %d, xEnd = %d, line = %d\n",xStart,xEnd,line);
-////printf("text = ");
-//  textLine = contents.at(line);
-//
-//  z = 0;
-//  x = 0;
-//  do {
-//      xc = x;
-//      ch = textLine->getChar(z);
-//      if (ch==127)
-//      {
-//       ch='?';
-//      }
-//      else
-//       if (ch !='\t' && (ch & 0x7f)< 0x20)
-//       {
-//         ch|=0x40;
-//       }
-//
-//      if (ch == '\t') {
-//          x += tabWidth - (x % tabWidth);
-//      } else {
-//          a = &attribs[textLine->getAttr(z)];
-//#ifdef QT_I18N
-//      if (iseucchar(ch)) {
-//	char buf[3];
-//	buf[0] = ch;
-//	buf[1] = textLine->getChar(++z);
-//	buf[2] = '\0';
-//	x += a->fm.width(buf);
-//      } else {
-//	x += a->fm.width(&ch,1);
-//      }
-//#else
-//          x += a->fm.width(&ch,1);
-//#endif
-//      }
-//      z++;
-//  } while (x <= xStart);
-//  zc = z - 1;
-//#ifdef QT_I18N
-//  if (iseucchar(textLine->getChar(zc))) {
-//    zc--;
-//  }
-//#endif
-//
-//  xs = xStart;
-//  attr = textLine->getRawAttr(zc);
-//  ch = textLine->getChar(zc);
-//  if (ch==127 || (ch !='\t' && ((ch & 0x7f)< 0x20)))
-//  {
-//        attr|=taSelected;
-//  }
-//
-//  while (x < xEnd) {
-//      nextAttr = textLine->getRawAttr(z);
-//      ch = textLine->getChar(z);
-//      /* non printable char handling */
-//      if (ch==127)
-//      {
-//        ch='?';
-//        nextAttr|=taSelected;
-//      }
-//      else
-//        if (ch !='\t' && ((ch & 0x7f)< 0x20))
-//        {
-//          ch|=0x40;
-//          nextAttr|=taSelected;
-//        }
-//
-//      if ((nextAttr ^ attr) & (taSelectMask | 256)) {
-//          paint.fillRect(xs - xStart,y,x - xs,fontHeight,colors[attr >> taShift]);
-//          xs = x;
-//          attr = nextAttr;
-//      }
-//      ch = textLine->getChar(z);
-//      if (ch == '\t') {
-//          x += tabWidth - (x % tabWidth);
-//      } else {
-//          a = &attribs[attr & taAttrMask];
-//#ifdef QT_I18N
-//      if (iseucchar(ch)) {
-// 	char buf[3];
-// 	buf[0] = ch;
-// 	buf[1] = textLine->getChar(++z);
-//	buf[2] = '\0';
-//	x += a->fm.width(buf);
-//      } else {
-//	x += a->fm.width(&ch,1);
-//      }
-//#else
-//          x += a->fm.width(&ch,1);
-//#endif
-//      }
-//      z++;
-//  }
-//
-//  paint.fillRect(xs - xStart,y,xEnd - xs,fontHeight,colors[attr >> taShift]);
-//
-//  y += fontAscent -1;
-//  attr = -1;
-//  aCnt = 0;
-//  xcp  = xc;
-//
-//  while (xc < xEnd) {
-//      ch = textLine->getChar(zc);
-//
-//      if (ch == '\t') {
-//          xc += tabWidth - (xc % tabWidth);
-//          if(aCnt!=0) {
-//              paint.drawText(xcp - xStart,y,aLB,aCnt);
-//              aCnt = 0;
-//          }
-//          xcp  = xc;
-//      }
-//
-//      else {
-//          nextAttr = textLine->getRawAttr(zc);
-//          /* non printable char handling */
-//          if (ch==127)
-//          {
-//            ch='?';
-//            nextAttr|=taSelected;
-//          }
-//          else
-//            if ((ch & 0x7f)< 0x20)
-//            {
-//              ch|=0x40;
-//              nextAttr|=taSelected;
-//            }
-//          if (nextAttr != attr) {
-//
-//              if(aCnt!=0) {
-//                  paint.drawText(xcp - xStart,y,aLB,aCnt);
-//                  aCnt = 0;
-//                  xcp  = xc;
-//              }
-//
-//              attr = nextAttr;
-//              a = &attribs[attr & taAttrMask];
-//              if (attr & taSelectMask) paint.setPen(a->selCol); else paint.setPen(a->col);
-//              paint.setFont(a->font);
-//          }
-//#ifdef QT_I18N
-//      if (iseucchar(ch)) {
-//         char buf[3];
-// 	buf[0] = ch;
-// 	buf[1] = textLine->getChar(++zc);
-// 	buf[2] = '\0';
-// 	paint.drawText(xc - xStart,y+1,buf);
-//	xc += a->fm.width(buf);
-//      } else {
-//        paint.drawText(xc - xStart,y,&ch,1);
-//	xc += a->fm.width(&ch,1);
-//      }
-//#else
-//          aLB[aCnt++] = ch;
-//          xc += a->fm.width(&ch,1);
-//
-//          if(aCnt == 256) {
-//              paint.drawText(xcp - xStart,y,aLB,aCnt);
-//              aCnt = 0;
-//              xcp  = xc;
-//          }
-//#endif
-//      }
-//      zc++;
-//  }
-//
-//  if(aCnt!=0) {
-//      paint.drawText(xcp - xStart,y,aLB,aCnt);
-//      aCnt = 0;
-//      xcp  = xc;
-//  }
-////printf("\n");
-//}
-
 //////////////////////////////////////////////////////////////////////////
-
+/*
 void KWriteDoc::printTextLine(QPainter &paint, int line, int xEnd, int y) {
   TextLine *textLine;
   int z, x;
@@ -2682,6 +2688,7 @@ void KWriteDoc::printTextLine(QPainter &paint, int line, int xEnd, int y) {
     paint.drawText(x, y, str);
   }
 }
+*/
 
 void KWriteDoc::setModified(bool m) {
   KWriteView *view;
@@ -3209,7 +3216,7 @@ void KWriteDoc::recordEnd(KWriteView *view, PointStruc &cursor, int flags) {
   view->updateCursor(cursor);
 
   optimizeSelection();
-  if (tagStart <= tagEnd) updateLines(tagStart,tagEnd,flags);
+  if (tagStart <= tagEnd) updateLines(tagStart,cursor.y,tagEnd,flags);
   setModified(true);
   newUndo();
 }
@@ -3249,7 +3256,7 @@ void KWriteDoc::doActionGroup(KWActionGroup *g, int flags) {
   }
   
   optimizeSelection();
-  if (tagStart <= tagEnd) updateLines(tagStart,tagEnd,flags);
+  if (tagStart <= tagEnd) updateLines(tagStart,-1,tagEnd,flags);
   setModified(true);
   newUndo();
 }
@@ -3518,7 +3525,7 @@ void KWriteDoc::newBracketMark(PointStruc &cursor, BracketMark &bm)
         x = 0;
       }
       if (textLine->getAttr(x) == attr) {
-        //try to find opposite bracked
+        //try to find opposite bracket
         ch = textLine->getChar(x);
         if (ch == bracket) count++; //same bracket : increase counter
         if (ch == opposite) {
